/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_BATCHNORM_H_
#define AIDGE_CORE_OPERATOR_BATCHNORM_H_

#include <array>
#include <memory>
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {

enum class BatchNormAttr { Epsilon, Momentum };

template <DimIdx_t DIM>
class BatchNorm_Op : public OperatorTensor,
                public Registrable<BatchNorm_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const BatchNorm_Op<DIM> &)>,
                public StaticAttributes<BatchNormAttr, float, float> {
public:
    static constexpr const char *Type = "BatchNorm";

    BatchNorm_Op() = delete;

    using Attributes_ = StaticAttributes<BatchNormAttr, float, float>;
    template <BatchNormAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr BatchNorm_Op(float epsilon, float momentum)
        : OperatorTensor(Type, 1, 4, 1),
          Attributes_(attr<BatchNormAttr::Epsilon>(epsilon),
                           attr<BatchNormAttr::Momentum>(momentum)) {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    BatchNorm_Op(const BatchNorm_Op<DIM>& op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        mImpl = op.mImpl ? Registrar<BatchNorm_Op<DIM>>::create(mOutputs[0]->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::BatchNorm_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<BatchNorm_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }


    void computeOutputDims() override final {
        // check inputs have been associated
        bool associated = true;
        for (IOIndex_t i = 0; i < nbInputs(); ++i) {
            associated &= !(getInput(i)->empty());
        }
        if (associated) {
            const DimSize_t nbChannels =  getInput(0)->dims()[1];
            for (std::size_t i = nbData(); i < nbInputs(); ++i) {
                if(getInput(i)->size() != nbChannels) {
                    // /!\ Input size should be handled BEFORE calling this function
                    // This should raise an error
                    getInput(i)->resize(std::array<DimSize_t, 1>({getInput(0)->dims()[1]}));
                }
            }
            mOutputs[0]->resize(getInput(0)->dims());
        }
    }

    void setBackend(const std::string &name) override {
        mImpl = Registrar<BatchNorm_Op<DIM>>::create(name)(*this);
        mOutputs[0]->setBackend(name);

        // FIXME: temporary workaround
        getInput(1)->setBackend(name);
        getInput(2)->setBackend(name);
        getInput(3)->setBackend(name);
        getInput(4)->setBackend(name);
    }

    static const std::vector<std::string> getInputsName() {
        return {"data_input", "scale", "shift", "mean", "variance"};
    }
    static const std::vector<std::string> getOutputsName() {
        return {"data_output"};
    }
};

template <DimSize_t DIM>
inline std::shared_ptr<Node> BatchNorm(const float epsilon = 1.0e-5F,
                                       const float momentum = 0.1F,
                                       const std::string& name = "") {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by BatchNorm, not supported");
    auto batchNorm = std::make_shared<Node>(std::make_shared<BatchNorm_Op<static_cast<DimIdx_t>(DIM)>>(epsilon, momentum), name);
    addProducer(batchNorm, 1, std::array<DimSize_t,0>({}), "scale");
    addProducer(batchNorm, 2, std::array<DimSize_t,0>({}), "shift");
    addProducer(batchNorm, 3, std::array<DimSize_t,0>({}), "batch_mean");
    addProducer(batchNorm, 4, std::array<DimSize_t,0>({}), "batch_variance");
    return batchNorm;
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::BatchNormAttr>::data[] = { "Epsilon", "Momentum" };
}

#endif //AIDGE_CORE_OPERATOR_BATCHNORM_H_