/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_CONV_H_
#define AIDGE_CORE_OPERATOR_CONV_H_

#include <array>
#include <cmath>
#include <cstddef>
#include <numeric>
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class ConvAttr { StrideDims, DilationDims, InChannels, OutChannels, KernelDims };

template <DimIdx_t DIM>
class Conv_Op : public OperatorTensor,
                public Registrable<Conv_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const Conv_Op<DIM> &)>,
                public StaticAttributes<ConvAttr, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>, DimSize_t,
                                       DimSize_t, std::array<DimSize_t, DIM>> {

public:
    static constexpr const char *Type = "Conv";

    Conv_Op() = delete;

    using Attributes_ = StaticAttributes<ConvAttr, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>,
                                             DimSize_t, DimSize_t, std::array<DimSize_t, DIM>>;
    template <ConvAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr Conv_Op(DimSize_t inChannels,
                      DimSize_t outChannels,
                      const std::array<DimSize_t, DIM> &kernelDims,
                      const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1))
        : OperatorTensor(Type, 1, 2, 1),
          Attributes_(attr<ConvAttr::StrideDims>(strideDims),
                      attr<ConvAttr::DilationDims>(dilationDims),
                      attr<ConvAttr::InChannels>(inChannels),
                      attr<ConvAttr::OutChannels>(outChannels),
                      attr<ConvAttr::KernelDims>(kernelDims)) {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Conv_Op(const Conv_Op<DIM>& op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        mImpl = op.mImpl ? Registrar<Conv_Op<DIM>>::create(mOutputs[0]->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Conv_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Conv_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    // std::shared_ptr<Conv_Op> clone() const override final {

    // }

    void computeOutputDims() override final {
        // check inputs have been associated
        bool associated = true;
        for (IOIndex_t i = 0; i < 3; ++i) {
            if (!getInput(i)) {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "Every input should be associated with a Tensor");
            }
            associated &= !(getInput(i)->empty());
        }
        if (associated) {
            std::array<DimSize_t, DIM + 2> outputDims{};
            const std::array<DimSize_t, DIM + 2> inputDims(getInput(0)->template dims<DIM+2>());

            for (std::size_t dim = 0; dim < this->template getAttr<ConvAttr::KernelDims>().size() ; ++dim) {
                const DimSize_t kernelExtent = this->template getAttr<ConvAttr::DilationDims>()[dim] *
                                                       (this->template getAttr<ConvAttr::KernelDims>()[dim] - 1) +
                                               1;

                outputDims[dim+2] = 1 + static_cast<DimSize_t>(
                        floor(static_cast<float>(inputDims[dim+2] - kernelExtent) /
                              static_cast<float>(this->template getAttr<ConvAttr::StrideDims>()[dim])));
            }

            outputDims[1] = this->template getAttr<ConvAttr::OutChannels>();
            outputDims[0] = inputDims[0];
            mOutputs[0]->resize(outputDims);
        }
    }


// std::vector<std::pair<std::size_t, std::vector<DimSize_t>>> computeReceptiveField(const std::size_t firstIdx, const std::vector<DimSize_t>& outputDims, const IOIndex_t outputIdx = 0) const override {
    //     if (outputIdx != 0) {
    //         AIDGE_THROW_OR_ABORT(std::runtime_error, "Conv_Op Operator has got only one output Tensor.");
    //     }
    //     if ((outputDims.size() == (DIM+2)) && outputDimsForwarded()) {
    //         // Offset
    //         const auto outputIdxDims = mOutput->getCoord(firstIdx);
    //         auto inputIdxDims = outputIdxDims; // batch idx is the same
    //         inputIdxDims[1] = 0; // each channel is used so start with the first one

    //         for (DimIdx_t i = 0; i < (DIM+2); ++i) {
    //             if (((outputDims[i] + outputIdxDims[i]) > mOutput->template dims<DIM+2>()[i]) || (outputDims[i] == 0)) {
    //                 AIDGE_THROW_OR_ABORT(std::runtime_error, "Given outputDim out of range for dimension %lu (%lu + %lu)", static_cast<std::size_t>(i), outputIdxDims[i], outputDims[i]);
    //             }
    //         }

    //         // padding is not a parameter of Conv_Op. It is handled in Pad_Op Operator
    //         // Input
    //         // same batch value, every input channel is used
    //         std::vector<DimSize_t> inputDims{outputDims[0], mInputs[0]->dims()[1]};
    //                     for (DimIdx_t i = 0; i < DIM; ++i) {
    //             inputDims.push_back((outputDims[2+static_cast<std::size_t>(i)] - 1)
    //                         * this->template getAttr<ConvAttr::StrideDims>()[static_cast<std::size_t>(i)]
    //                         + 1
    //                         + (this->template getAttr<ConvAttr::KernelDims>()[static_cast<std::size_t>(i)] - 1)
    //                         * this->template getAttr<ConvAttr::DilationDims>()[static_cast<std::size_t>(i)]);
    //             inputIdxDims[2+i] *= this->template getAttr<ConvAttr::StrideDims>()[static_cast<std::size_t>(i)];
    //         }

    //         // Weight
    //         // same output value, every input channel is used
    //         std::vector<DimSize_t> weightDims{outputDims[0], mInputs[0]->dims()[1]};
    //         weightDims.insert(weightDims.end(), this->template getAttr<ConvAttr::KernelDims>()[0], this->template getAttr<ConvAttr::KernelDims>()[static_cast<std::size_t>(DIM)]);
    //         std::vector<DimSize_t> weightIdxDims = std::vector<DimSize_t>(DIM+2, 0);
    //         weightIdxDims[0] = outputIdxDims[1];

    //         // Bias
    //         const std::vector<DimSize_t> biasDims{outputDims[0]};
    //         const std::vector<DimSize_t> biasIdxDims{outputIdxDims[1]};

    //         // Result
    //         std::vector<std::pair<std::size_t, std::vector<DimSize_t>>> res;
    //         res.push_back(std::pair<std::size_t, std::vector<DimSize_t>>(mInputs[0]->getIdx(inputIdxDims), inputDims));
    //         res.push_back(std::pair<std::size_t, std::vector<DimSize_t>>(mInputs[1]->getIdx(weightIdxDims), weightDims));
    //         res.push_back(std::pair<std::size_t, std::vector<DimSize_t>>(mInputs[2]->getIdx(biasIdxDims), biasDims));
    //         return res;
    //     }
    //     AIDGE_THROW_OR_ABORT(std::runtime_error, "Given outputDim out of range or output dim not forwarded yet.");
    // }

    void setBackend(const std::string &name) override {
        mImpl = Registrar<Conv_Op<DIM>>::create(name)(*this);
        mOutputs[0]->setBackend(name);

        // FIXME: temporary workaround
        getInput(1)->setBackend(name);
        getInput(2)->setBackend(name);
    }

    static const std::vector<std::string> getInputsName(){
        return {"data_input", "weight", "bias"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Conv(DimSize_t inChannels,
                                  DimSize_t outChannels,
                                  const std::array<DimSize_t, DIM> &kernelDims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    auto conv = std::make_shared<Node>(std::make_shared<Conv_Op<static_cast<DimIdx_t>(DIM)>>(inChannels, outChannels, kernelDims, strideDims, dilationDims), name);
    // addProducer(conv, 1, append(append(kernel_dims, in_channels), out_channels), "w");
    addProducer(conv, 1, append(outChannels, append(inChannels, kernelDims)), "w");
    addProducer(conv, 2, std::array<DimSize_t, 1>({outChannels}), "b");
    return conv;
}

// helper with C-style array instead of std::array for kernel_dims to allow automatic template DIM deduction
template <DimSize_t DIM>
inline std::shared_ptr<Node> Conv(
    DimSize_t inChannels,
    DimSize_t outChannels,
    DimSize_t const (&kernelDims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    return Conv(inChannels, outChannels, to_array(kernelDims), name, strideDims, dilationDims);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ConvAttr>::data[] = {
    "StrideDims",
    "DilationDims",
    "InChannels",
    "OutChannels",
    "KernelDims"
};
}

#endif /* AIDGE_CORE_OPERATOR_CONV_H_ */